/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// vid_x.c -- X Window System video driver

#include "quakedef.h"
#include "xquake.h"
#include "d_local.h"
#include "quake.xpm"

static int scrnum;
static qboolean doShm;
static Colormap x_cmap;
static GC x_gc;
static Visual *x_vis;
static XVisualInfo *x_visinfo;

static int x_shmeventtype;
static qboolean oktodraw = false;

int XShmQueryExtension(Display *);
int XShmGetEventBase(Display *);

int current_framebuffer;
static XImage *x_framebuffer[2] = {0, 0};
static XShmSegmentInfo x_shminfo[2];

static qboolean verbose = false;

static byte current_palette[768];

static long X11_buffersize;

int vid_surfcachesize;
void *vid_surfcache;

Atom wm_delete_window_atom;

viddef_t vid; // global video state

unsigned short d_8to16table[256]; // for software renderer

cvar_t vid_mode = {"vid_mode","0",false};

static XF86VidModeModeInfo init_vidmode, game_vidmode;

//==========================================================================
//
// Palette stuff
//
//==========================================================================
typedef unsigned short PIXEL16;
typedef unsigned long PIXEL24;
static PIXEL16 st2d_8to16table[256];
static PIXEL24 st2d_8to24table[256];
static int shiftmask_fl = 0;
static long r_shift,g_shift,b_shift;
static unsigned long r_mask,g_mask,b_mask;

void shiftmask_init (void)
{
	unsigned int x;

	r_mask = x_vis->red_mask;
	g_mask = x_vis->green_mask;
	b_mask = x_vis->blue_mask;

	if (r_mask > (1 << 31) || g_mask > (1 << 31) || b_mask > (1 << 31))
		Sys_Error ("shiftmask_init: XGetVisualInfo returned bogus rgb mask\n");

	for (r_shift = -8, x = 1; x < r_mask; x = x<<1)
		r_shift++;
	for (g_shift = -8, x = 1; x < g_mask; x = x<<1)
		g_shift++;
	for (b_shift = -8, x = 1; x < b_mask; x = x<<1)
		b_shift++;

	shiftmask_fl = 1;
}

PIXEL16 xlib_rgb16(int r,int g,int b)
{
	PIXEL16 p;

	if(shiftmask_fl == 0) 
		shiftmask_init();

	p = 0;

	if (r_shift > 0) 
		p = (r<<(r_shift)) & r_mask;
	else if (r_shift < 0) 
		p = (r>>(-r_shift)) & r_mask;
	else 
		p |= (r & r_mask);

	if (g_shift > 0) 
		p |= (g<<(g_shift)) & g_mask;
	else if (g_shift < 0) 
		p |= (g>>(-g_shift)) & g_mask;
	else 
		p |= (g & g_mask);

	if (b_shift > 0) 
		p |= (b<<(b_shift)) & b_mask;
	else if (b_shift < 0) 
		p |= (b>>(-b_shift)) & b_mask;
	else 
		p |= (b & b_mask);

	return p;
}

PIXEL24 xlib_rgb24(int r,int g,int b)
{
	PIXEL24 p;

	if(shiftmask_fl == 0) 
		shiftmask_init();

	p = 0;

	if (r_shift > 0) 
		p = (r<<(r_shift)) & r_mask;
	else if (r_shift < 0) 
		p = (r>>(-r_shift)) & r_mask;
	else 
		p |= (r & r_mask);

	if (g_shift > 0) 
		p |= (g<<(g_shift)) & g_mask;
	else if (g_shift < 0) 
		p |= (g>>(-g_shift)) & g_mask;
	else 
		p |= (g & g_mask);

	if (b_shift > 0) 
		p |= (b<<(b_shift)) & b_mask;
	else if (b_shift < 0) 
		p |= (b>>(-b_shift)) & b_mask;
	else 
		p |= (b & b_mask);

	return p;
}

//==========================================================================
//
// Framebuffer stuff
//
//==========================================================================
void st2_fixup (XImage *framebuf, int x, int y, int width, int height)
{
//	int xi;
	int yi;
	byte *src;
	PIXEL16 *dest;
	register int count, n;

	if (x < 0 || y < 0)
		return;

	for (yi = y; yi < (y + height); yi++) 
	{
		src = (byte *)&framebuf->data [yi * framebuf->bytes_per_line];

		// Duff's Device
		count = width;
		n = (count + 7) / 8;
		dest = ((PIXEL16 *)src) + x + width - 1;
		src += x + width - 1;

		switch (count % 8) {
		case 0:	do {	*dest-- = st2d_8to16table[*src--];
		case 7:			*dest-- = st2d_8to16table[*src--];
		case 6:			*dest-- = st2d_8to16table[*src--];
		case 5:			*dest-- = st2d_8to16table[*src--];
		case 4:			*dest-- = st2d_8to16table[*src--];
		case 3:			*dest-- = st2d_8to16table[*src--];
		case 2:			*dest-- = st2d_8to16table[*src--];
		case 1:			*dest-- = st2d_8to16table[*src--];
				} while (--n > 0);
		}

//		for(xi = (x + width - 1); xi >= x; xi--) 
//		{
//			dest[xi] = st2d_8to16table[src[xi]];
//		}
	}
}

void st3_fixup (XImage *framebuf, int x, int y, int width, int height)
{
//	int xi;
	int yi;
	byte *src;
	PIXEL24 *dest;
	register int count, n;

	if (x < 0 || y < 0)
		return;

	for (yi = y; yi < (y + height); yi++) {
		src = (byte *)&framebuf->data [yi * framebuf->bytes_per_line];

		// Duff's Device
		count = width;
		n = (count + 7) / 8;
		dest = ((PIXEL24 *)src) + x + width - 1;
		src += x + width - 1;

		switch (count % 8) {
		case 0:	do {	*dest-- = st2d_8to24table[*src--];
		case 7:			*dest-- = st2d_8to24table[*src--];
		case 6:			*dest-- = st2d_8to24table[*src--];
		case 5:			*dest-- = st2d_8to24table[*src--];
		case 4:			*dest-- = st2d_8to24table[*src--];
		case 3:			*dest-- = st2d_8to24table[*src--];
		case 2:			*dest-- = st2d_8to24table[*src--];
		case 1:			*dest-- = st2d_8to24table[*src--];
				} while (--n > 0);
		}

//		for(xi = (x + width - 1); xi >= x; xi--) 
//		{
//			dest[xi] = st2d_8to16table[src[xi]];
//		}
	}
}

/*
================
ResetFrameBuffer
================
*/
void ResetFrameBuffer (void)
{
	int mem;
	int pwidth;

	if (x_framebuffer[0])
	{
		free(x_framebuffer[0]->data);
		free(x_framebuffer[0]);
	}

	if (d_pzbuffer)
	{
		D_FlushCaches ();
		free (d_pzbuffer);
	}

// alloc an extra line in case we want to wrap, and allocate the z-buffer
	X11_buffersize = vid.width * vid.height * sizeof (*d_pzbuffer);

	vid_surfcachesize = D_SurfaceCacheForRes (vid.width, vid.height);

	X11_buffersize += vid_surfcachesize;

	d_pzbuffer = malloc (X11_buffersize);
	d_pzsize = X11_buffersize;

	if (d_pzbuffer == NULL)
		Sys_Error ("ResetFrameBuffer: Not enough memory for video mode\n");

	vid_surfcache = (byte *) d_pzbuffer + vid.width * vid.height * sizeof (*d_pzbuffer);

	D_InitCaches(vid_surfcache, vid_surfcachesize);

	pwidth = x_visinfo->depth / 8;
	if (pwidth == 3) 
		pwidth = 4;
	mem = ((vid.width * pwidth + 7) & ~7) * vid.height;

	x_framebuffer[0] = XCreateImage (
		x_disp,
		x_vis,
		x_visinfo->depth,
		ZPixmap,
		0,
		malloc(mem),
		vid.width, vid.height,
		32,
		0
	);

	if (!x_framebuffer[0])
		Sys_Error("ResetFrameBuffer: XCreateImage failed\n");

	vid.buffer = (byte *) (x_framebuffer[0]);
	vid.conbuffer = vid.buffer;
}

/*
================
ResetSharedFrameBuffers
================
*/
void ResetSharedFrameBuffers (void)
{
	int size;
	int key;
	int minsize = getpagesize ();
	int frm;

	if (d_pzbuffer)
	{
		D_FlushCaches ();
		free (d_pzbuffer);
	}

// alloc an extra line in case we want to wrap, and allocate the z-buffer
	X11_buffersize = vid.width * vid.height * sizeof (*d_pzbuffer);

	vid_surfcachesize = D_SurfaceCacheForRes (vid.width, vid.height);

	X11_buffersize += vid_surfcachesize;

	d_pzbuffer = malloc (X11_buffersize);
	d_pzsize = X11_buffersize;

	if (d_pzbuffer == NULL)
		Sys_Error ("ResetSharedFrameBuffers: Not enough memory for video mode\n");

	vid_surfcache = (byte *) d_pzbuffer + vid.width * vid.height * sizeof (*d_pzbuffer);

	D_InitCaches(vid_surfcache, vid_surfcachesize);

	for (frm=0 ; frm<2 ; frm++)
	{

		// free up old frame buffer memory
		if (x_framebuffer[frm])
		{
			XShmDetach(x_disp, &x_shminfo[frm]);
			free(x_framebuffer[frm]);
			shmdt(x_shminfo[frm].shmaddr);
		}

		// create the image
		x_framebuffer[frm] = XShmCreateImage (
			x_disp,
			x_vis,
			x_visinfo->depth,
			ZPixmap,
			0,
			&x_shminfo[frm],
			vid.width,
			vid.height 
		);

		// grab shared memory
		size = x_framebuffer[frm]->bytes_per_line * x_framebuffer[frm]->height;
		if (size < minsize)
			Sys_Error("ResetSharedFrameBuffers: Window must use at least %d bytes\n", minsize);

		key = random();
		x_shminfo[frm].shmid = shmget((key_t)key, size, IPC_CREAT | 0777);
		if (x_shminfo[frm].shmid == -1)
			Sys_Error("ResetSharedFrameBuffers: Could not get any shared memory\n");

		// attach to the shared memory segment
		x_shminfo[frm].shmaddr = (void *)shmat(x_shminfo[frm].shmid, 0, 0);

		if (verbose)
		{
			printf("Video:\n");
			printf("   shared memory id=%d, addr=0x%lx\n", x_shminfo[frm].shmid, (long)x_shminfo[frm].shmaddr);
		}

		x_framebuffer[frm]->data = x_shminfo[frm].shmaddr;

		// get the X server to attach to it
		if (!XShmAttach(x_disp, &x_shminfo[frm]))
			Sys_Error("ResetSharedFrameBuffers: XShmAttach() failed\n");

		XSync(x_disp, False);
		shmctl(x_shminfo[frm].shmid, IPC_RMID, 0);
	}
}


//
// direct draw software compatability stuff
//
/*
================
D_BeginDirectRect
================
*/
void D_BeginDirectRect (int x, int y, byte *pbitmap, int width, int height)
{
// direct drawing of the "accessing disk" icon isn't supported under Linux
}

/*
================
D_EndDirectRect
================
*/
void D_EndDirectRect (int x, int y, int width, int height)
{
// direct drawing of the "accessing disk" icon isn't supported under Linux
}


/*
===========
HandleEvents

getting and handle events
===========
*/
void HandleEvents (void)
{
	XEvent x_event;

	qboolean dowarp = false;

	if (!x_disp)
		return;

	while (XPending(x_disp)) 
	{
		XNextEvent(x_disp, &x_event);

		switch (x_event.type) 
		{
			case KeyPress: // key pressed
			case KeyRelease: // key released
				Key_Event(XLateKey(&x_event.xkey), x_event.type == KeyPress);
				break;

			case MotionNotify: // mouse moved
				if (mouse_grab_active) 
				{
					if (dga_mouse_active) 
					{
						mouse_x += x_event.xmotion.x_root;
						mouse_y += x_event.xmotion.y_root;
					} 
					else 
					{
						if (!x_event.xmotion.send_event)
						{
							mouse_x += x_event.xmotion.x - p_mouse_x;
							mouse_y += x_event.xmotion.y - p_mouse_y;
							if (abs(vid.width/2 - x_event.xmotion.x) > vid.width / 4 || abs(vid.height/2 - x_event.xmotion.y) > vid.height / 4)
								dowarp = true;
						}
						p_mouse_x = x_event.xmotion.x;
						p_mouse_y = x_event.xmotion.y;
					}
				}
				break;

			case ButtonPress: // mouse button pressed
			case ButtonRelease: // mouse button released
				switch (x_event.xbutton.button)
				{
					case 1:
						Key_Event (K_MOUSE1, x_event.type == ButtonPress);
						break;

					case 2:
						Key_Event (K_MOUSE3, x_event.type == ButtonPress);
						break;

					case 3:
						Key_Event (K_MOUSE2, x_event.type == ButtonPress);
						break;

					case 4:
						Key_Event (K_MWHEELUP, x_event.type == ButtonPress);
						break;

					case 5:
						Key_Event (K_MWHEELDOWN, x_event.type == ButtonPress);
						break;

					case 6:
						Key_Event (K_MOUSE4, x_event.type == ButtonPress);
						break;

					case 7:
						Key_Event (K_MOUSE5, x_event.type == ButtonPress);
						break;

					case 8:
						Key_Event (K_MOUSE6, x_event.type == ButtonPress);
						break;

					case 9:
						Key_Event (K_MOUSE7, x_event.type == ButtonPress);
						break;

					case 10:
						Key_Event (K_MOUSE8, x_event.type == ButtonPress);
						break;
				}
				break;

			case CreateNotify: // window created
				win_x = x_event.xcreatewindow.x;
				win_y = x_event.xcreatewindow.y;
				break;

			case ConfigureNotify: // window changed size/location
				win_x = x_event.xconfigure.x;
				win_y = x_event.xconfigure.y;
				win_width = x_event.xconfigure.width;
				win_height = x_event.xconfigure.height;
				win_notify = true;
				break;

			case DestroyNotify: // window has been destroyed
				Sys_Quit();
				break; 

			case ClientMessage: // window manager messages
				if ((x_event.xclient.format == 32) && ((unsigned int)x_event.xclient.data.l[0] == wm_delete_window_atom))
					Sys_Quit();
				break; 

			case MapNotify: // window restored
				vid_hidden = false;
				if(vidmode_fullscreen)
				{
					// set our video mode
					XF86VidModeSwitchToMode(x_disp, scrnum, &game_vidmode);

					// Move the viewport to top left
					XF86VidModeSetViewPort(x_disp, scrnum, 0, 0);
				}
				break;

			case UnmapNotify: // window iconified/rolledup/whatever
				vid_hidden = true;
				if(vidmode_fullscreen)
				{
					// set our video mode
					XF86VidModeSwitchToMode(x_disp, scrnum, &init_vidmode);
				}
				break;

			case FocusIn: // window is now the input focus
				vid_activewindow = true;
				break; 

			case FocusOut: // window is no longer the input focus
				vid_activewindow = false;
				break; 

			case EnterNotify: // mouse entered window
				vid_notifywindow = true;
				break;

			case LeaveNotify: // mouse left window
				vid_notifywindow = false;
				break;

			default:
				if (doShm && x_event.type == x_shmeventtype)
					oktodraw = true;
		}
	}

	if (dowarp) 
	{
		/* move the mouse to the window center again */
		p_mouse_x = vid.width / 2;
		p_mouse_y = vid.height / 2;
		XWarpPointer(x_disp, None, x_win, 0, 0, 0, 0, p_mouse_x, p_mouse_y);
	}
}

/*
===========
Sys_SendKeyEvents
===========
*/
void Sys_SendKeyEvents (void)
{
	static qboolean active = true;

	// do some checks, if windowed
	if (!vidmode_fullscreen)
	{
		// enable/disable sound, set/restore gamma and grab/ungrab keyb
		// on focus gain/loss
		if (!vid_hidden && vid_activewindow)
		{
			if (!active)
			{
				S_UnblockSound ();
				S_ClearBuffer ();
				IN_GrabKeyboard();
				active = true;
			}
		}
		else
		{
			if (active)
			{
				S_BlockSound ();
				S_ClearBuffer ();
				IN_UngrabKeyboard();
				active = false;
			}
		}

		// handle the mouse state when windowed if that's changed
		if (key_dest != key_console /* && !mouse_grab_active */ && vid_activewindow)
		{
			IN_GrabMouse ();
		}
		else if ( /* mouse_grab_active && */ key_dest == key_console) 
		{
			IN_UngrabMouse ();
		}
	}

	HandleEvents ();
}

/*
===========
VID_Update

flushes the given rectangles from the view buffer to the screen
===========
*/
void VID_Update (vrect_t *rects)
{
	// if the window changes dimension, skip this frame
	if (!vidmode_fullscreen)
	{
		// check for resize
		if (win_notify)
		{
			win_notify = false;

			if (win_width < 320)
				win_width = 320;
			if (win_height < 200)
				win_height = 200;

			vid.width = win_width;
			vid.height = win_height;

			if (doShm)
				ResetSharedFrameBuffers();
			else
				ResetFrameBuffer();

			vid.rowbytes = x_framebuffer[0]->bytes_per_line;
			vid.buffer = (pixel_t *)x_framebuffer[current_framebuffer]->data;
			vid.conbuffer = vid.buffer;
			vid.conrowbytes = vid.rowbytes;
			vid.conwidth = vid.width;
			vid.conheight = vid.height;
			vid.aspect = ((float)vid.height / (float)vid.width) * (320.0 / 240.0); // FIXME

			vid.recalc_refdef = true; // force a surface cache flush

			return;
		}
	}

	// force full update if not 8bit
	if (x_visinfo->depth != 8) 
	{
		scr_fullupdate = 0;
	}

	if (doShm)
	{
		while (rects)
		{
			if (x_visinfo->depth == 16)
				st2_fixup( x_framebuffer[current_framebuffer], rects->x, rects->y, rects->width, rects->height);
			else if (x_visinfo->depth == 24)
				st3_fixup( x_framebuffer[current_framebuffer], rects->x, rects->y, rects->width, rects->height);

			if (!XShmPutImage(x_disp, x_win, x_gc, x_framebuffer[current_framebuffer], rects->x, rects->y, rects->x, rects->y, rects->width, rects->height, True))
					Sys_Error("VID_Update: XShmPutImage failed\n");

			oktodraw = false;

			while (!oktodraw) 
				HandleEvents();

			rects = rects->pnext;
		}
		current_framebuffer = !current_framebuffer;
		vid.buffer = (pixel_t *)x_framebuffer[current_framebuffer]->data;
		vid.conbuffer = vid.buffer;
		XSync (x_disp, False);
	}
	else
	{
		while (rects)
		{
			if (x_visinfo->depth == 16)
				st2_fixup( x_framebuffer[current_framebuffer], rects->x, rects->y, rects->width, rects->height);
			else if (x_visinfo->depth == 24)
				st3_fixup( x_framebuffer[current_framebuffer], rects->x, rects->y, rects->width, rects->height);

			XPutImage (x_disp, x_win, x_gc, x_framebuffer[0], rects->x, rects->y, rects->x, rects->y, rects->width, rects->height);
			rects = rects->pnext;
		}
		XSync (x_disp, False);
	}
}

/*
===============
VID_SetPalette
===============
*/
void VID_SetPalette (byte *palette)
{

	int i;
	XColor colors[256];

	for(i = 0; i < 256; i++) 
	{
		st2d_8to16table[i]= xlib_rgb16(palette[i * 3], palette[i * 3 + 1], palette[i * 3 + 2]);
		st2d_8to24table[i]= xlib_rgb24(palette[i * 3], palette[i * 3 + 1], palette[i * 3 + 2]);
	}

	if (x_visinfo->class == PseudoColor && x_visinfo->depth == 8)
	{
		if (palette != current_palette)
			memcpy(current_palette, palette, 768);
		for (i = 0; i < 256; i++)
		{
			colors[i].pixel = i;
			colors[i].flags = DoRed | DoGreen | DoBlue;
			colors[i].red = palette[i * 3] * 257;
			colors[i].green = palette[i * 3 + 1] * 257;
			colors[i].blue = palette[i * 3 + 2] * 257;
		}
		XStoreColors(x_disp, x_cmap, colors, 256);
	}
}

/*
===============
VID_ShiftPalette
===============
*/
void VID_ShiftPalette (byte *p)
{
	VID_SetPalette (p);
}

/*
===============
VID_Init

Called at startup to set up translation tables, takes 256 8 bit RGB values
the palette data will go away after the call, so it must be copied off if
the video driver will need it again
===============
*/
void VID_Init(byte *palette)
{
	int i;
	XVisualInfo template;
	int num_visuals;
	int template_mask;
	XSetWindowAttributes attr;
	XClassHint *clshints;
	XWMHints *wmhints;
	XSizeHints *szhints;
	unsigned long mask;
	Window root;
	qboolean fullscreen = true;
	qboolean vidmode_ext = false;
	qboolean VidMode = false;
	int MajorVersion, MinorVersion;

	Cvar_RegisterVariable (&vid_mode);

// set vid parameters
	vid.width = 640;
	vid.height = 480;
	vid.maxwarpwidth = WARP_WIDTH;
	vid.maxwarpheight = WARP_HEIGHT;
	vid.numpages = 2;
	vid.colormap = host_colormap;
	vid.fullbright = 256 - LittleLong (*((int *)vid.colormap + 2048));

// open the display
	x_disp = XOpenDisplay(NULL);
	if (!x_disp)
	{
		if (getenv("DISPLAY"))
			Sys_Error("VID_Init: Could not open display [%s]\n", getenv("DISPLAY"));
		else
			Sys_Error("VID_Init: Could not open local display\n");
	}

// get video mode list
	MajorVersion = MinorVersion = 0;
	VidMode = XF86VidModeQueryVersion(x_disp, &MajorVersion, &MinorVersion);
//	if (XF86VidModeQueryVersion(x_disp, &MajorVersion, &MinorVersion)) 
	if (COM_CheckParm("-novidmode"))
	{
		Con_Warning ("XFree86 VidMode extension disabled at command line\n");
	}
	else if (VidMode)
	{
		Con_Printf("XFree86 VidMode extension version %d.%d found\n", MajorVersion, MinorVersion);
		vidmode_ext = true;
	}
	else
	{
		Con_Warning ("XFree86 VidMode extension not supported\n");
	}

// for debugging only
	XSynchronize(x_disp, True);

// specify a visual id
	template_mask = 0;
	if ((i = COM_CheckParm("-visualid")))
	{
		if (i >= com_argc-1)
			Sys_Error("VID_Init: -visualid <id#>\n");
		template.visualid = atoi(com_argv[i+1]);
		template_mask = VisualIDMask;
	}
// if not specified, use default visual
	else
	{
		int screen;

		screen = XDefaultScreen(x_disp);
		template.visualid = XVisualIDFromVisual(XDefaultVisual(x_disp, screen));
		template_mask = VisualIDMask;
	}

// pick a visual- warn if more than one was available
	x_visinfo = XGetVisualInfo(x_disp, template_mask, &template, &num_visuals);
	x_vis = x_visinfo->visual;
	scrnum = x_visinfo->screen;
	root = XRootWindow(x_disp, scrnum);
	if (num_visuals > 1)
	{
		printf("Found more than one visual id at depth %d:\n", template.depth);
		for (i=0 ; i<num_visuals ; i++)
			printf("   -visualid %d\n", (int)(x_visinfo[i].visualid));
	}
	else if (num_visuals == 0)
	{
		if (template_mask == VisualIDMask)
			Sys_Error("VID_Init: Bad visual id %d\n", template.visualid);
		else
			Sys_Error("VID_Init: No visuals at depth %d\n", template.depth);
	}

// check for command-line parameters
	verbose = COM_CheckParm ("-verbose");
	if (verbose)
	{
		printf("Using visualid %d:\n", (int)(x_visinfo->visualid));
		printf("   screen %d\n", x_visinfo->screen);
		printf("   red_mask 0x%x\n", (int)(x_visinfo->red_mask));
		printf("   green_mask 0x%x\n", (int)(x_visinfo->green_mask));
		printf("   blue_mask 0x%x\n", (int)(x_visinfo->blue_mask));
		printf("   colormap_size %d\n", x_visinfo->colormap_size);
		printf("   bits_per_rgb %d\n", x_visinfo->bits_per_rgb);
	}

// check for command-line size parameters
	if ((i = COM_CheckParm("-width"))) 
	{
		if (i >= com_argc-1)
			Sys_Error("VID_Init: -width <width>\n");

		vid.width = atoi(com_argv[i+1]);
		if (!vid.width)
			Sys_Error("VID_Init: Bad width\n");
	}
	if ((i = COM_CheckParm("-height"))) 
	{
		if (i >= com_argc-1)
			Sys_Error("VID_Init: -height <height>\n");

		vid.height = atoi(com_argv[i+1]);
		if (!vid.height)
			Sys_Error("VID_Init: Bad height\n");
	}

// check for command-line video parameters
	if (COM_CheckParm("-current"))
	{
		vid.width = DisplayWidth (x_disp, scrnum);
		vid.height = DisplayHeight (x_disp, scrnum);
	}
	else if (COM_CheckParm("-window"))
	{
		fullscreen = false;
	}

// set the vidmode to the requested width/height if possible 
	if (vidmode_ext) 
	{
		int x, y, dist;
		int best_dist = 9999999;
		int num_modes;
		XF86VidModeModeInfo **modes;
		XF86VidModeModeInfo *mode = NULL;

		XF86VidModeGetAllModeLines(x_disp, scrnum, &num_modes, &modes);

		// are we going fullscreen?  If so, let's change video mode
		if (fullscreen) 
		{
			for (i = 0; i < num_modes; i++) 
			{
				if (vid.width > modes[i]->hdisplay || vid.height > modes[i]->vdisplay)
					continue;

				x = vid.width - modes[i]->hdisplay;
				y = vid.height - modes[i]->vdisplay;
				dist = (x * x) + (y * y);

				if (dist < best_dist) 
				{
					best_dist = dist;
					mode = modes[i];
				}
			}

			if (mode)
			{
				// change to the mode
				XF86VidModeSwitchToMode(x_disp, scrnum, mode);
				memcpy(&game_vidmode, mode, sizeof(game_vidmode)); // save game vidmode
				memcpy(&init_vidmode, modes[0], sizeof(init_vidmode)); // save system initial vidmode
				vidmode_fullscreen = true;

				// Move the viewport to top left
				XF86VidModeSetViewPort(x_disp, scrnum, 0, 0);
				Con_DPrintf ("VID_Init: set fullscreen mode at %dx%d\n", vid.width, vid.height); 
			} 
			else
			{
//				fullscreen = false; // eh?
				Con_DPrintf ("VID_Init: unable to set fullscreen mode at %dx%d\n", vid.width, vid.height); 
			}

			free(modes);
		}
	}

// window attributes
	mask = CWBackPixel | CWColormap | CWEventMask;
	attr.background_pixel = 0;
	attr.colormap = XCreateColormap(x_disp, root, x_vis, AllocNone);
	attr.event_mask = X_MASK;

// setup attributes for main window
	if (vidmode_fullscreen) 
	{
		mask |= CWSaveUnder | CWBackingStore | CWOverrideRedirect;
		attr.override_redirect = True;
		attr.backing_store = NotUseful;
		attr.save_under = False;
	} 
	else
	{
		mask |= CWBorderPixel;
		attr.border_pixel = 0;
	}

// create the main window
	x_win = XCreateWindow (
		x_disp,
		root,
		0, 0,	// x, y
		vid.width, vid.height,
		0, // borderwidth
		x_visinfo->depth,
		InputOutput,
		x_vis,
		mask,
		&attr 
	);

	wmhints = XAllocWMHints();
	if(XpmCreatePixmapFromData(x_disp, x_win, quake_xpm, &wmhints->icon_pixmap, &wmhints->icon_mask, NULL) == XpmSuccess)
		wmhints->flags |= IconPixmapHint | IconMaskHint;

	clshints = XAllocClassHint();
	clshints->res_name = strdup("SourceQuake");
	clshints->res_class = strdup("SourceQuake");

	szhints = XAllocSizeHints();
	if(vidmode_fullscreen)
	{
		szhints->min_width = szhints->max_width = vid.width;
		szhints->min_height = szhints->max_height = vid.height;
		szhints->flags |= PMinSize | PMaxSize;
	}

	XmbSetWMProperties(x_disp, x_win, "SourceQuake", "SourceQuake", (char **) com_argv, com_argc, szhints, wmhints, clshints);
	XFree(clshints);
	XFree(wmhints);
	XFree(szhints);

//	if (!vidmode_fullscreen)
//		XStoreName(x_disp, x_win, "SourceQuake");

	if (x_visinfo->depth == 8)
	{
		// create and upload the palette
		if (x_visinfo->class == PseudoColor)
		{
			x_cmap = XCreateColormap(x_disp, x_win, x_vis, AllocAll);
			VID_SetPalette(palette);
			XSetWindowColormap(x_disp, x_win, x_cmap);
		}
	}

// create the GC
	{
		XGCValues xgcvalues;
		int valuemask = GCGraphicsExposures;
		xgcvalues.graphics_exposures = False;
		x_gc = XCreateGC(x_disp, x_win, valuemask, &xgcvalues);
	}

// map the window
	XMapWindow(x_disp, x_win);

// making the close button on a window do the right thing
// seems to involve this mess, sigh...
	wm_delete_window_atom = XInternAtom(x_disp, "WM_DELETE_WINDOW", false);
	XSetWMProtocols(x_disp, x_win, &wm_delete_window_atom, 1);

	if (vidmode_fullscreen) 
	{
		XMoveWindow(x_disp, x_win, 0, 0);
		XRaiseWindow(x_disp, x_win);
		XWarpPointer(x_disp, None, x_win, 0, 0, 0, 0, 0, 0);
		XFlush(x_disp);

		// Move the viewport to top left
		XF86VidModeSetViewPort(x_disp, scrnum, 0, 0);
	}

	XFlush(x_disp);

// wait for first exposure event
	{
		XEvent x_event;
		do {
			XNextEvent(x_disp, &x_event);
			if (x_event.type == Expose && !x_event.xexpose.count)
				oktodraw = true;
		} while (!oktodraw);
	}

// now safe to draw

// even if MITSHM is available, make sure it's a local connection
	if (XShmQueryExtension(x_disp))
	{
		char *displayname;
		doShm = true;
		displayname = (char *) getenv("DISPLAY");
		if (displayname)
		{
			char *d = displayname;
			while (*d && (*d != ':')) 
				d++;
			if (*d) 
				*d = 0;
			if (!(!strcasecmp(displayname, "unix") || !*displayname))
				doShm = false;
		}
	}

	if (doShm)
	{
		x_shmeventtype = XShmGetEventBase(x_disp) + ShmCompletion;
		ResetSharedFrameBuffers();
	}
	else
	{
		ResetFrameBuffer();
	}

	vid_activewindow = true;
	vid_hidden = false;
	vid_notifywindow = true;

	current_framebuffer = 0;
	vid.rowbytes = x_framebuffer[0]->bytes_per_line;
	vid.buffer = (pixel_t *)x_framebuffer[0]->data;
	vid.direct = 0;
	vid.conbuffer = (pixel_t *)x_framebuffer[0]->data;
	vid.conrowbytes = vid.rowbytes;
	vid.conwidth = vid.width;
	vid.conheight = vid.height;
	vid.aspect = ((float)vid.height / (float)vid.width) * (320.0 / 240.0);

	InitSig(); // trap signals

// for debugging only
//	XSynchronize(x_disp, False);

	Con_SafePrintf ("Video mode %dx%d %s initialized\n", vid.width, vid.height, vidmode_fullscreen ? "fullscreen" : "windowed");
}

/*
===============
VID_Shutdown

called at shutdown
===============
*/
void VID_Shutdown (void)
{
	if (x_disp) 
	{
		if (x_win)
			XDestroyWindow(x_disp, x_win);

		if (vidmode_fullscreen)
			XF86VidModeSwitchToMode(x_disp, scrnum, &init_vidmode);

		XCloseDisplay (x_disp);
	}

	vidmode_fullscreen = false;
	x_disp = NULL;
	x_win = 0;
}

/*
===============
VID_LockBuffer
===============
*/
void VID_LockBuffer (void)
{
}

/*
===============
VID_UnlockBuffer
===============
*/
void VID_UnlockBuffer (void)
{
}

